@@ -3,7 +3,7 @@ Sphinx Notion Builder.
 """
 
 import json
-from functools import singledispatch
+from functools import singledispatchmethod
 from typing import Any
 
 from beartype import beartype
@@ -45,36 +45,7 @@ from ultimate_notion.obj_api.core import GenericObject
 from ultimate_notion.obj_api.enums import BGColor, CodeLang
 from ultimate_notion.rich_text import Text, text
 
-
-def _process_list_item_recursively(node: nodes.list_item) -> UnoBulletedItem:
-    """
-    Recursively process a list item node and return a BulletedItem.
-    """
-    paragraph = node.children[0]
-    assert isinstance(paragraph, nodes.paragraph)
-    rich_text = _create_rich_text_from_children(node=paragraph)
-    block = UnoBulletedItem(text="placeholder")
-    block.rich_text = rich_text
-
-    bullet_only_msg = (
-        "The only thing Notion supports within a bullet list is a bullet list"
-    )
-
-    for child in node.children[1:]:
-        assert isinstance(child, nodes.bullet_list), bullet_only_msg
-        for nested_list_item in child.children:
-            assert isinstance(nested_list_item, nodes.list_item), (
-                bullet_only_msg
-            )
-            nested_block = _process_list_item_recursively(
-                node=nested_list_item,
-            )
-            # Remove pyright ignore once we have
-            # https://github.com/ultimate-notion/ultimate-notion/issues/94.
-            # pylint: disable=line-too-long
-            block.obj_ref.value.children.append(nested_block.obj_ref)  # pyright: ignore[reportUnknownMemberType]
-
-    return block
+type _BlockTree = dict[tuple[NotionObject[Any], int], "_BlockTree"]
 
 
 def _create_rich_text_from_children(*, node: nodes.Element) -> Text:
@@ -165,272 +136,6 @@ def _cell_source_node(entry: nodes.Node) -> nodes.paragraph:
     return combined
 
 
-@singledispatch
-def _process_node_to_blocks(
-    node: nodes.Element,
-    *,
-    section_level: int,
-) -> list[NotionObject[Any]]:  # pragma: no cover
-    """
-    Required function for ``singledispatch``.
-    """
-    del section_level
-    raise NotImplementedError(node)
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.table, *, section_level: int) -> list[NotionObject[Any]]:
-    """Process rST table nodes by creating Notion Table blocks.
-
-    This implementation delegates small branches to helpers which keeps
-    the function body linear and easier to reason about.
-    """
-    del section_level
-
-    n_cols, header_row, body_rows = _extract_table_structure(node=node)
-
-    n_rows = 1 + len(body_rows) if header_row else len(body_rows)
-    table = UnoTable(n_rows=n_rows, n_cols=n_cols, header_row=bool(header_row))
-
-    row_idx = 0
-    if header_row is not None:
-        for col_idx, entry in enumerate(iterable=header_row.children):
-            source = _cell_source_node(entry=entry)
-            table[row_idx, col_idx] = _create_rich_text_from_children(
-                node=source
-            )
-        row_idx += 1
-
-    for body_row in body_rows:
-        for col_idx, entry in enumerate(iterable=body_row.children):
-            source = _cell_source_node(entry=entry)
-            table[row_idx, col_idx] = _create_rich_text_from_children(
-                node=source
-            )
-        row_idx += 1
-
-    return [table]
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.paragraph, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process paragraph nodes by creating Notion Paragraph blocks.
-    """
-    del section_level
-    rich_text = _create_rich_text_from_children(node=node)
-    paragraph_block = UnoParagraph(text="")
-    paragraph_block.rich_text = rich_text
-    return [paragraph_block]
-
-
-@_process_node_to_blocks.register
-def _(
-    node: nodes.block_quote, *, section_level: int
-) -> list[NotionObject[Any]]:
-    """
-    Process block quote nodes by creating Notion Quote blocks.
-    """
-    del section_level
-    rich_text = _create_rich_text_from_children(node=node)
-    quote_block = UnoQuote(text="")
-    quote_block.rich_text = rich_text
-    return [quote_block]
-
-
-@_process_node_to_blocks.register
-def _(
-    node: nodes.literal_block, *, section_level: int
-) -> list[NotionObject[Any]]:
-    """
-    Process literal block nodes by creating Notion Code blocks.
-    """
-    del section_level
-    code_text = _create_rich_text_from_children(node=node)
-    pygments_lang = node.get(key="language", failobj="")
-    language = _map_pygments_to_notion_language(
-        pygments_lang=pygments_lang,
-    )
-    code_block = UnoCode(text=code_text, language=language)
-    # By default, the code block has a color set (DEFAULT) which means
-    # that there is no syntax highlighting.
-    # See https://github.com/ultimate-notion/ultimate-notion/issues/93.
-    for rich_text in code_text.rich_texts:  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
-        del rich_text.obj_ref.annotations  # pyright: ignore[reportUnknownMemberType]
-    code_block.rich_text = code_text
-    return [code_block]
-
-
-@_process_node_to_blocks.register
-def _(
-    node: nodes.bullet_list,
-    *,
-    section_level: int,
-) -> list[NotionObject[Any]]:
-    """
-    Process bullet list nodes by creating Notion BulletedItem blocks.
-    """
-    del section_level
-    return [
-        _process_list_item_recursively(node=list_item)
-        for list_item in node.children
-        if isinstance(list_item, nodes.list_item)
-    ]
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.topic, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process topic nodes, specifically for table of contents.
-    """
-    del section_level  # Not used for topics
-    # Later, we can support `.. topic::` directives, likely as
-    # a callout with no icon.
-    assert "contents" in node["classes"]
-    return [UnoTableOfContents()]
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.compound, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process Sphinx ``toctree`` nodes.
-    """
-    del node
-    del section_level
-    # There are no specific Notion blocks for ``toctree`` nodes.
-    # We need to support ``toctree`` in ``index.rst``.
-    # Just ignore it.
-    return []
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.title, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process title nodes by creating appropriate Notion heading blocks.
-    """
-    rich_text = _create_rich_text_from_children(node=node)
-
-    heading_levels: dict[int, type[UnoHeading[Any]]] = {
-        1: UnoHeading1,
-        2: UnoHeading2,
-        3: UnoHeading3,
-    }
-    heading_cls = heading_levels[section_level]
-    block = heading_cls(text="")
-
-    block.rich_text = rich_text
-    return [block]
-
-
-def _create_admonition_callout(
-    node: nodes.Element,
-    *,
-    emoji: str,
-    background_color: BGColor,
-) -> list[NotionObject[Any]]:
-    """Create a Notion Callout block for admonition nodes.
-
-    The first child (typically a paragraph) becomes the callout text,
-    and any remaining children become nested blocks within the callout.
-    """
-    block = UnoCallout(
-        text="", icon=Emoji(emoji=emoji), color=background_color
-    )
-
-    # Use the first child as the callout text
-    first_child = node.children[0]
-    if isinstance(first_child, nodes.paragraph):
-        rich_text = _create_rich_text_from_children(node=first_child)
-        block.rich_text = rich_text
-        # Process remaining children as nested blocks
-        children_to_process = node.children[1:]
-    else:
-        # If first child is not a paragraph, use empty text
-        block.rich_text = Text.from_plain_text(text="")
-        # Process all children as nested blocks (including the first)
-        children_to_process = node.children
-
-    # Process children as nested blocks
-    for child in children_to_process:
-        for child_block in list(
-            _process_node_to_blocks(
-                child,
-                section_level=1,
-            )
-        ):
-            # Add nested blocks as children to the callout
-            # Remove pyright ignore once we have
-            # https://github.com/ultimate-notion/ultimate-notion/issues/94.
-            block.obj_ref.value.children.append(child_block.obj_ref)  # pyright: ignore[reportUnknownMemberType]
-
-    return [block]
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.note, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process note admonition nodes by creating Notion Callout blocks.
-    """
-    del section_level
-    return _create_admonition_callout(
-        node=node,
-        emoji="📝",
-        background_color=BGColor.BLUE,
-    )
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.warning, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process warning admonition nodes by creating Notion Callout blocks.
-    """
-    del section_level
-    return _create_admonition_callout(
-        node=node,
-        emoji="⚠️",
-        background_color=BGColor.YELLOW,
-    )
-
-
-@_process_node_to_blocks.register
-def _(node: nodes.tip, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process tip admonition nodes by creating Notion Callout blocks.
-    """
-    del section_level
-    return _create_admonition_callout(
-        node=node,
-        emoji="💡",
-        background_color=BGColor.GREEN,
-    )
-
-
-@_process_node_to_blocks.register
-def _(node: CollapseNode, *, section_level: int) -> list[NotionObject[Any]]:
-    """
-    Process collapse nodes by creating Notion ToggleItem blocks.
-    """
-    del section_level
-
-    children_to_process = node.children
-    title_text = node.attributes["label"]
-    toggle_block = UnoToggleItem(text=title_text)
-
-    for child in children_to_process:
-        for child_block in list(
-            _process_node_to_blocks(
-                child,
-                section_level=1,
-            )
-        ):
-            # Add nested blocks as children to the toggle
-            # Remove pyright ignore once we have
-            # https://github.com/ultimate-notion/ultimate-notion/issues/94.
-            toggle_block.obj_ref.value.children.append(child_block.obj_ref)  # pyright: ignore[reportUnknownMemberType]
-
-    return [toggle_block]
-
-
 def _map_pygments_to_notion_language(*, pygments_lang: str) -> CodeLang:
     """
     Map ``Pygments`` language names to Notion CodeLang ``enum`` values.
@@ -546,19 +251,395 @@ class NotionTranslator(NodeVisitor):
         """
         del builder
         super().__init__(document=document)
-        self._blocks: list[NotionObject[Any]] = []
+        self._block_tree: _BlockTree = {}
         self.body: str
         self._section_level = 0
 
+    def _add_block_to_tree(
+        self,
+        block: NotionObject[Any],
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """Add a block to the block tree.
+
+        First has to find the parent in the tree recursively.
+        """
+        block_key = (block, id(block))
+        if not parent_path:
+            self._block_tree[block_key] = {}
+            return
+
+        current_node = self._block_tree[parent_path[0]]
+
+        for parent_key in parent_path[1:]:
+            current_node = current_node[parent_key]
+        current_node[block_key] = {}
+
+    def _process_list_item_recursively(
+        self,
+        *,
+        node: nodes.list_item,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Recursively process a list item node and return a BulletedItem.
+        """
+        paragraph = node.children[0]
+        assert isinstance(paragraph, nodes.paragraph)
+        rich_text = _create_rich_text_from_children(node=paragraph)
+        block = UnoBulletedItem(text="placeholder")
+        block.rich_text = rich_text
+        self._add_block_to_tree(
+            block=block,
+            parent_path=parent_path,
+        )
+
+        bullet_only_msg = (
+            "The only thing Notion supports within a bullet list is a "
+            "bullet list"
+        )
+        assert isinstance(node, nodes.list_item)
+
+        for child in node.children[1:]:
+            assert isinstance(child, nodes.bullet_list), bullet_only_msg
+            for nested_list_item in child.children:
+                assert isinstance(nested_list_item, nodes.list_item), (
+                    bullet_only_msg
+                )
+                self._process_list_item_recursively(
+                    node=nested_list_item,
+                    parent_path=[*parent_path, (block, id(block))],
+                )
+
+    @singledispatchmethod
+    def _process_node_to_blocks(  # pylint: disable=no-self-use
+        self,
+        node: nodes.Element,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:  # pragma: no cover
+        """
+        Required function for ``singledispatch``.
+        """
+        del section_level
+        del parent_path
+        raise NotImplementedError(node)
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.table,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """Process rST table nodes by creating Notion Table blocks.
+
+        This implementation delegates small branches to helpers which
+        keeps the function body linear and easier to reason about.
+        """
+        del section_level
+
+        n_cols, header_row, body_rows = _extract_table_structure(node=node)
+
+        n_rows = 1 + len(body_rows) if header_row else len(body_rows)
+        table = UnoTable(
+            n_rows=n_rows, n_cols=n_cols, header_row=bool(header_row)
+        )
+
+        row_idx = 0
+        if header_row is not None:
+            for col_idx, entry in enumerate(iterable=header_row.children):
+                source = _cell_source_node(entry=entry)
+                table[row_idx, col_idx] = _create_rich_text_from_children(
+                    node=source
+                )
+            row_idx += 1
+
+        for body_row in body_rows:
+            for col_idx, entry in enumerate(iterable=body_row.children):
+                source = _cell_source_node(entry=entry)
+                table[row_idx, col_idx] = _create_rich_text_from_children(
+                    node=source
+                )
+            row_idx += 1
+
+        self._add_block_to_tree(block=table, parent_path=parent_path)
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.paragraph,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process paragraph nodes by creating Notion Paragraph blocks.
+        """
+        del section_level
+        rich_text = _create_rich_text_from_children(node=node)
+        paragraph_block = UnoParagraph(text="")
+        paragraph_block.rich_text = rich_text
+        self._add_block_to_tree(block=paragraph_block, parent_path=parent_path)
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.block_quote,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process block quote nodes by creating Notion Quote blocks.
+        """
+        del section_level
+        rich_text = _create_rich_text_from_children(node=node)
+        quote_block = UnoQuote(text="")
+        quote_block.rich_text = rich_text
+        self._add_block_to_tree(block=quote_block, parent_path=parent_path)
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.literal_block,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process literal block nodes by creating Notion Code blocks.
+        """
+        del section_level
+        code_text = _create_rich_text_from_children(node=node)
+        pygments_lang = node.get(key="language", failobj="")
+        language = _map_pygments_to_notion_language(
+            pygments_lang=pygments_lang,
+        )
+        code_block = UnoCode(text=code_text, language=language)
+        # By default, the code block has a color set (DEFAULT) which means
+        # that there is no syntax highlighting.
+        # See https://github.com/ultimate-notion/ultimate-notion/issues/93.
+        for rich_text in code_text.rich_texts:  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
+            del rich_text.obj_ref.annotations  # pyright: ignore[reportUnknownMemberType]
+        code_block.rich_text = code_text
+        self._add_block_to_tree(block=code_block, parent_path=parent_path)
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.bullet_list,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process bullet list nodes by creating Notion BulletedItem blocks.
+        """
+        del section_level
+        bullet_only_msg = (
+            "The only thing Notion supports within a bullet list is a "
+            "bullet list"
+        )
+        for list_item in node.children:
+            assert isinstance(list_item, nodes.list_item), bullet_only_msg
+            self._process_list_item_recursively(
+                node=list_item,
+                parent_path=parent_path,
+            )
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.topic,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process topic nodes, specifically for table of contents.
+        """
+        del section_level  # Not used for topics
+        # Later, we can support `.. topic::` directives, likely as
+        # a callout with no icon.
+        assert "contents" in node["classes"]
+        toc_block = UnoTableOfContents()
+        self._add_block_to_tree(block=toc_block, parent_path=parent_path)
+
+    @_process_node_to_blocks.register
+    def _(  # pylint: disable=no-self-use
+        self,
+        node: nodes.compound,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process Sphinx ``toctree`` nodes.
+        """
+        del node
+        del section_level
+        del parent_path
+        # There are no specific Notion blocks for ``toctree`` nodes.
+        # We need to support ``toctree`` in ``index.rst``.
+        # Just ignore it.
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.title,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process title nodes by creating appropriate Notion heading blocks.
+        """
+        rich_text = _create_rich_text_from_children(node=node)
+
+        heading_levels: dict[int, type[UnoHeading[Any]]] = {
+            1: UnoHeading1,
+            2: UnoHeading2,
+            3: UnoHeading3,
+        }
+        heading_cls = heading_levels[section_level]
+        block = heading_cls(text="")
+
+        block.rich_text = rich_text
+        self._add_block_to_tree(block=block, parent_path=parent_path)
+
+    def _create_admonition_callout(
+        self,
+        node: nodes.Element,
+        *,
+        emoji: str,
+        background_color: BGColor,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """Create a Notion Callout block for admonition nodes.
+
+        The first child (typically a paragraph) becomes the callout
+        text, and any remaining children become nested blocks within the
+        callout.
+        """
+        block = UnoCallout(
+            text="", icon=Emoji(emoji=emoji), color=background_color
+        )
+
+        self._add_block_to_tree(block=block, parent_path=parent_path)
+        # Use the first child as the callout text
+        first_child = node.children[0]
+        if isinstance(first_child, nodes.paragraph):
+            rich_text = _create_rich_text_from_children(node=first_child)
+            block.rich_text = rich_text
+            # Process remaining children as nested blocks
+            children_to_process = node.children[1:]
+        else:
+            # If first child is not a paragraph, use empty text
+            block.rich_text = Text.from_plain_text(text="")
+            # Process all children as nested blocks (including the first)
+            children_to_process = node.children
+
+        # Process children as nested blocks
+        for child in children_to_process:
+            self._process_node_to_blocks(
+                child,
+                section_level=1,
+                parent_path=[*parent_path, (block, id(block))],
+            )
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.note,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process note admonition nodes by creating Notion Callout blocks.
+        """
+        del section_level
+        self._create_admonition_callout(
+            node=node,
+            emoji="📝",
+            background_color=BGColor.BLUE,
+            parent_path=parent_path,
+        )
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.warning,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process warning admonition nodes by creating Notion Callout blocks.
+        """
+        del section_level
+        self._create_admonition_callout(
+            node=node,
+            emoji="⚠️",
+            background_color=BGColor.YELLOW,
+            parent_path=parent_path,
+        )
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: nodes.tip,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process tip admonition nodes by creating Notion Callout blocks.
+        """
+        del section_level
+        self._create_admonition_callout(
+            node=node,
+            emoji="💡",
+            background_color=BGColor.GREEN,
+            parent_path=parent_path,
+        )
+
+    @_process_node_to_blocks.register
+    def _(
+        self,
+        node: CollapseNode,
+        *,
+        section_level: int,
+        parent_path: list[tuple[NotionObject[Any], int]],
+    ) -> None:
+        """
+        Process collapse nodes by creating Notion ToggleItem blocks.
+        """
+        del section_level
+
+        title_text = node.attributes["label"]
+        toggle_block = UnoToggleItem(text=title_text)
+        self._add_block_to_tree(block=toggle_block, parent_path=parent_path)
+
+        for child in node.children:
+            self._process_node_to_blocks(
+                child,
+                section_level=1,
+                parent_path=[*parent_path, (toggle_block, id(toggle_block))],
+            )
+
     def visit_title(self, node: nodes.Element) -> None:
         """
         Handle title nodes by creating appropriate Notion heading blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -580,66 +661,71 @@ class NotionTranslator(NodeVisitor):
         """
         Handle paragraph nodes by creating Notion Paragraph blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
+
         raise nodes.SkipNode
 
     def visit_block_quote(self, node: nodes.Element) -> None:
         """
         Handle block quote nodes by creating Notion Quote blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
+
         raise nodes.SkipNode
 
     def visit_literal_block(self, node: nodes.Element) -> None:
         """
         Handle literal block nodes by creating Notion Code blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
+
         raise nodes.SkipNode
 
     def visit_bullet_list(self, node: nodes.Element) -> None:
         """
         Handle bullet list nodes by processing each list item.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
+
         raise nodes.SkipNode
 
     def visit_topic(self, node: nodes.Element) -> None:
         """
         Handle topic nodes, specifically for table of contents.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
+
         raise nodes.SkipNode
 
     def visit_note(self, node: nodes.Element) -> None:
         """
         Handle note admonition nodes by creating Notion Callout blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -647,11 +733,11 @@ class NotionTranslator(NodeVisitor):
         """
         Handle warning admonition nodes by creating Notion Callout blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -659,11 +745,11 @@ class NotionTranslator(NodeVisitor):
         """
         Handle tip admonition nodes by creating Notion Callout blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -671,11 +757,11 @@ class NotionTranslator(NodeVisitor):
         """
         Handle table nodes by creating Notion Table blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -683,11 +769,11 @@ class NotionTranslator(NodeVisitor):
         """
         Handle compound admonition nodes by creating a table of contents block.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -695,11 +781,11 @@ class NotionTranslator(NodeVisitor):
         """
         Handle collapse nodes by creating Notion ToggleItem blocks.
         """
-        blocks = _process_node_to_blocks(
+        self._process_node_to_blocks(
             node,
             section_level=self._section_level,
+            parent_path=[],
         )
-        self._blocks.extend(blocks)
 
         raise nodes.SkipNode
 
@@ -710,20 +796,37 @@ class NotionTranslator(NodeVisitor):
         assert self
         del node
 
-    def depart_document(self, node: nodes.Element) -> None:
+    def _convert_block_tree_to_json(
+        self,
+        # `beartype` does not support recursive types, so we need to use a
+        # simpler type.
+        block_tree: dict[tuple[NotionObject[Any], int], Any],
+    ) -> list[dict[str, Any]]:
         """
-        Output collected blocks as JSON at document end.
+        Convert the block tree to a JSON-serializable format, ignoring IDs from
+        tuples.
         """
-        del node
-        dumped_blocks: list[dict[str, Any]] = []
-        for block in self._blocks:
+        result: list[dict[str, Any]] = []
+        for (block, _), subtree in block_tree.items():
             obj_ref = block.obj_ref
             assert isinstance(obj_ref, GenericObject)
-            dumped_block = obj_ref.serialize_for_api()
-            dumped_blocks.append(dumped_block)
+            dumped_structure = {
+                "block": obj_ref.serialize_for_api(),
+                "children": self._convert_block_tree_to_json(
+                    block_tree=subtree
+                ),
+            }
+            result.append(dumped_structure)
+        return result
+
+    def depart_document(self, node: nodes.Element) -> None:
+        """
+        Output collected block tree as JSON at document end.
+        """
+        del node
 
         json_output = json.dumps(
-            obj=dumped_blocks,
+            obj=self._convert_block_tree_to_json(block_tree=self._block_tree),
             indent=2,
             ensure_ascii=False,
         )
