@@ -61,6 +61,34 @@ def _create_code_block_without_annotations(
     return code_block
 
 
+def _reconstruct_nested_structure(
+    items: list[dict[str, Any]],
+) -> list[dict[str, Any]]:
+    """
+    Reconstruct the nested structure of the nodes from the graph produced.
+    """
+    result: list[dict[str, Any]] = []
+    for item in items:
+        block = item["block"]
+        block_type = block["type"]
+
+        # Handle blocks that can have children
+        if block_type in {
+            "paragraph",
+            "callout",
+            "bulleted_list_item",
+            "toggle",
+        }:
+            children = item.get("children", [])
+            nested_blocks = _reconstruct_nested_structure(
+                items=children,
+            )
+            block[block_type]["children"] = nested_blocks
+
+        result.append(block)
+    return result
+
+
 def _assert_rst_converts_to_notion_objects(
     rst_content: str,
     expected_objects: list[NotionObject[Any]],
@@ -91,6 +119,10 @@ def _assert_rst_converts_to_notion_objects(
     with output_file.open() as f:
         generated_json: list[dict[str, Any]] = json.load(fp=f)
 
+    generated_json_un_flattened = _reconstruct_nested_structure(
+        items=generated_json
+    )
+
     expected_json: list[dict[str, Any]] = []
     for notion_object in expected_objects:
         obj_ref = notion_object.obj_ref
@@ -98,7 +130,10 @@ def _assert_rst_converts_to_notion_objects(
         dumped_block = obj_ref.serialize_for_api()
         expected_json.append(dumped_block)
 
-    assert generated_json == expected_json, (generated_json, expected_json)
+    assert generated_json_un_flattened == expected_json, (
+        generated_json_un_flattened,
+        expected_json,
+    )
 
 
 def test_single_paragraph(
